"""
Definitions of constraints
"""

import pysmt


def binary_temporal_constraint(t_1, disjunctive_distance, t_2):
    """
    Difference between t_2 and t_1 is within one of the disjunctive intervals
    :param t_1:
    :param disjunctive_distance:
    :param t_2:
    :return:
    """
    difference = pysmt.shortcuts.Minus(t_2, t_1)
    constraint = pysmt.shortcuts.Or([
        pysmt.shortcuts.And(pysmt.shortcuts.GE(difference, pysmt.shortcuts.Real(dd[0])),
                            pysmt.shortcuts.LE(difference, pysmt.shortcuts.Real(dd[1])))
        for dd in disjunctive_distance
    ])
    return constraint


def unary_temporal_constaint(t_p, disjunctive_distance):
    """
    The abolute time of tp is within one of the disjunctive intervals
    :param t_p:
    :param disjunctive_distance:
    :return:
    """
    constraint = pysmt.shortcuts.Or([
        pysmt.shortcuts.And(pysmt.shortcuts.GE(t_p, pysmt.shortcuts.Real(dd[0])),
                            pysmt.shortcuts.LE(t_p, pysmt.shortcuts.Real(dd[1])))
        for dd in disjunctive_distance
    ])
    return constraint


def join_constraint(t_join, joined_times):
    """
    A join step must be equal to one of the preceding timepoints
    :param t_join:
    :param joined_times:
    :return:
    """
    constraint = pysmt.shortcuts.Or([
        pysmt.shortcuts.Equals(t_join, t_j)
        for t_j in joined_times
    ])
    return constraint


def time_points_happen_once_constraint(timepoint_vars, happenings):
    """
    Each time point is equal to at least one happening
    :param timepoint_vars:
    :param happenings:
    :return: constraint
    """
    constraint = pysmt.shortcuts.And([
        pysmt.shortcuts.Or([
            pysmt.shortcuts.Equals(t, h)
            for h in happenings
        ])
        for _, t in timepoint_vars.items()
    ])
    return constraint

def duration_constraint(start, end, duration):
    #constraint = pysmt.shortcuts.Equals(duration,
    #                              pysmt.shortcuts.Minus(end, start))
    constraint = pysmt.shortcuts.And(pysmt.shortcuts.LE(duration,
                                                            pysmt.shortcuts.Minus(end, start)),
                                     pysmt.shortcuts.GE(duration,
                                                             pysmt.shortcuts.Minus(end, start)))

    return constraint

def anytime_before(t1, t2):
    return [[0,t_inf]]



def determine_time_constraint(source_uri, disjunctive_distance, sink_uri):
    t1 = timepoint_vars[source_uri]
    t2 = timepoint_vars[sink_uri]
    return binary_temporal_constraint(t1, disjunctive_distance, t2)


"""
Test a hand-coded temporal problem
"""
import pysmt.shortcuts
import paml_check.paml_check as pc
from paml_check.constraints import binary_temporal_constraint, join_constraint, \
    unary_temporal_constaint


def test_ludox_check_direct():
    """
    Hand coded temporal problem that is satisfiable
    :return:
    """
    timepoints = [
        "t_0",
        "t_F",
        "t_P1_s",
        "t_P2_s",
        "t_P1_e",
        "t_P2_e",
        "t_M",
        "t_MA_s",
        "t_MA_e",
        "t_N"
    ]
    t_inf = 10000.0
    t_epsilon = 0.0001
    # num_happenings = 10

    timepoint_vars = {t: pysmt.shortcuts.Symbol(t, pysmt.shortcuts.REAL)
                      for t in timepoints}

    timepoint_var_domains = [pysmt.shortcuts.And(pysmt.shortcuts.GE(t, pysmt.shortcuts.Real(0.0)),
                                                 pysmt.shortcuts.LE(t, pysmt.shortcuts.Real(t_inf)))
                             for _, t in timepoint_vars.items()]

    time_constraints = [
        binary_temporal_constraint(timepoint_vars["t_0"], [
                                   [0, 0]], timepoint_vars["t_F"]),
        binary_temporal_constraint(timepoint_vars["t_F"], [
                                   [0, t_inf]], timepoint_vars["t_P1_s"]),
        binary_temporal_constraint(timepoint_vars["t_F"], [
                                   [0, t_inf]], timepoint_vars["t_P2_s"]),
        binary_temporal_constraint(timepoint_vars["t_P1_s"], [
                                   [t_epsilon, t_inf]], timepoint_vars["t_P1_e"]),
        binary_temporal_constraint(timepoint_vars["t_P2_s"], [
                                   [t_epsilon, t_inf]], timepoint_vars["t_P2_e"]),
        binary_temporal_constraint(timepoint_vars["t_P1_e"], [
                                   [0, t_inf]], timepoint_vars["t_M"]),
        binary_temporal_constraint(timepoint_vars["t_P2_e"], [
                                   [0, t_inf]], timepoint_vars["t_M"]),
        binary_temporal_constraint(timepoint_vars["t_M"], [
                                   [t_epsilon, t_inf]], timepoint_vars["t_MA_s"]),
        binary_temporal_constraint(timepoint_vars["t_MA_s"], [
                                   [t_epsilon, t_inf]], timepoint_vars["t_MA_e"]),
        binary_temporal_constraint(timepoint_vars["t_MA_e"], [
                                   [0, 0]], timepoint_vars["t_N"]),
    ]

    join_constraints = [
        join_constraint(timepoint_vars["t_M"],
                        [
                            timepoint_vars["t_P2_e"],
                            timepoint_vars["t_P1_e"]
        ])
    ]

    durations = [
        binary_temporal_constraint(timepoint_vars["t_P1_s"], [
                                   [3, 3]], timepoint_vars["t_P1_e"]),
        binary_temporal_constraint(timepoint_vars["t_P2_s"], [
                                   [3, 3]], timepoint_vars["t_P2_e"]),
        binary_temporal_constraint(timepoint_vars["t_MA_s"], [
                                   [10, 10]], timepoint_vars["t_MA_e"]),
        binary_temporal_constraint(timepoint_vars["t_0"], [
                                   [0, 100]], timepoint_vars["t_N"])
    ]

    events = [
        unary_temporal_constaint(timepoint_vars["t_0"], [[1, 3]]),
        unary_temporal_constaint(timepoint_vars["t_N"], [[15, 16]])
    ]

    # happenings = [pysmt.shortcuts.Symbol(
    #    f"h_{i}", pysmt.shortcuts.REAL) for i in range(num_happenings)]

    # happening_timepoint_mappings = time_points_happen_once_constraint(
    #    timepoint_vars, happenings)

    given_constraints = pysmt.shortcuts.And(
        timepoint_var_domains + time_constraints + join_constraints)
    hand_coded_constraints = pysmt.shortcuts.And(durations + events)
    formula = pysmt.shortcuts.And(
        given_constraints,
        hand_coded_constraints,
    #    happening_timepoint_mappings
    )
    result = pc.check(formula)
    if result:
        #timeline = Timeline(result, timepoint_vars, happenings)
        # print(timeline)
        print("SAT")
    else:
        print("UNSAT")

    assert result



import os
import sbol3
import paml
import operator

from paml_check.activity_graph import ActivityGraph
import paml_check.paml_check as pc

paml_spec = "https://raw.githubusercontent.com/SD2E/paml/time/paml/paml.ttl"
        

# junk code to print out the results in a slightly easier to read output
def print_debug(result, graph):
    def make_entry(variable, activity, uri, value, prefix = ""):
        v = value
        if activity.start.identity == variable.identity:
            s = f"{prefix}S {activity.identity} : {value}"
        elif activity.end.identity == variable.identity:
            s = f"{prefix}E {activity.identity} : {value}"
        elif activity.duration.identity == variable.identity:
            s = f"{prefix}D {uri} : {value}"
        else:
            s = f"{prefix}ERR {uri} : {value}"
        return (v, s)
    nodes = []
    protcols = []
    for (node, value) in result:
        uri = node.symbol_name()
        v = (float)(value.constant_value())
        if uri in graph.nodes:
            variable = graph.nodes[uri]
            activity = variable.get_parent()
            nodes.append(make_entry(variable, activity, uri, v))
        else:
            variable = graph.doc.find(uri)
            activity = variable.get_parent()
            protcols.append(make_entry(variable, activity, uri, v, "PROTOCOL "))

    print("--- Nodes ---")
    for k in sorted(nodes, key=operator.itemgetter(0)):
        print(k[1])
    print("--- Protocol ---")
    for k in sorted(protcols, key=operator.itemgetter(0)):
        print(k[1])


def generate_and_test_constraints(paml_file):
    doc = sbol3.Document()
    doc.read(paml_file, 'ttl')
    graph = ActivityGraph(doc)

    formula = graph.generate_constraints()
    result = pc.check(formula)
    if result:
        graph.add_result(doc, result)
        graph.compute_durations(doc)
        graph.print_debug()
        print_debug(result, graph)
        print("SAT")
    else:
        print("UNSAT")
    assert result


def test_generate_timed_constraints():
    paml_file = os.path.join(os.getcwd(), 'resources/paml', 'igem_ludox_time_draft.ttl')
    generate_and_test_constraints(paml_file)


def test_generate_untimed_constraints():
    paml_file = os.path.join(os.getcwd(), 'resources/paml', 'igem_ludox_draft.ttl')
    generate_and_test_constraints(paml_file)


class Timeline():
    def __init__(self, model, timepoint_vars, happenings):
        self.happening_timepoints = {}
        self.time_variables = {}

        for var, value in result:
            if value.is_real_constant():
                if str(value) in self.time_variables:
                    if var in happenings:
                        self.time_variables[str(value)]["happenings"] = { "time" : value.constant_value(), "timepoints" : []}
        for var, value in result:
            if value.is_real_constant():
                if str(var) in timepoint_vars:
                    self.happening_timepoints["var"] = { "time" : value.constant_value(), "timepoints" : []}


"""
Setup script
"""
import os
from setuptools import setup, find_packages

# read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

def _post_install():
    print("Installing z3...")
    os.system("pysmt-install --z3 --confirm-agreement")
    os.system("export PYSMT_CYTHON=0")
    # PYSMT_CYTHON = 0


setup(name='paml_check',
      version='0.1',
      description='PAML Checker',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/SD2E/paml-check',
      author='Dan Bryce',
      author_email='dbryce@sift.net',
      license='MIT',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      install_requires=[
          # "paml" This requires that paml have a valid package name
          "pysmt",
          "sbol3"
      ],
      tests_require=["pytest"],
      zip_safe=False
      )

_post_install()
